/**
* ap.c - P2PWNC access point module
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*
* In addition, as a special exception, the copyright holders
* give permission to link the code of this program with the
* OpenSSL toolkit (or with modified versions of OpenSSL that use the
* same license as OpenSSL, version 0.9.8a), and distribute linked
* combinations including the two.  You must obey the GNU General Public License
* in all respects for all of the code used other than OpenSSL.  If you modify
* this file, you may extend this exception to your version of the
* file, but you are not obligated to do so.  If you do not wish to
* do so, delete this exception statement from your version.
*/

#include "ap.h"

PWNC_CLIENT *PWNC_CLIENT_new(unsigned long ipaddr) {
	PWNC_CLIENT *client = NULL;
	ALLOC_AND_CLEAN(client, sizeof(PWNC_CLIENT), PWNC_CLIENT*);
	client->client_ip = ipaddr;
	client->client_cert = PWNC_CERT_new();
	client->rcpt = NULL;
//	client->rcpt = PWNC_RCPT_new();
	return client;
}

unsigned long pwnc_client_hash(const PWNC_CLIENT *to_hash) {
	return to_hash->client_ip;
}

int pwnc_client_cmp(const PWNC_CLIENT *arg1, const PWNC_CLIENT *arg2) {
	return ( (int)(arg1->client_ip - arg2->client_ip));
}
//initializes message queue
int init_mq() {
	mq_id = msgget(mq_key, 0600);
	if (mq_id == -1) {
		fprintf(stderr, "init_mq: Error creating message queue\n");
	}
	return mq_id;
}

//sends command to message queue to be processed by the other process
int send_cmd(unsigned char *cmd_text) {
	//FW_CMD cmd;
	MSG_BUF *msg_to_send;

	int mlen = strlen(cmd_text); //length of the message text
	ALLOC_AND_CLEAN(msg_to_send, sizeof(MSG_BUF) + mlen, MSG_BUF*); //allocate memory for msgbuf
	msg_to_send->mtype = MTYPE_FW_CMD; //message type: iptables command
	memcpy(&(msg_to_send->mtext), cmd_text, mlen); //copy command string to cmd->command field

	//now we can send the message to the queue
	if (msgsnd(mq_id, msg_to_send, mlen, 0) == -1) {
		fprintf(stderr, "send_cmd: Error sending message to queue\n");
		return -1;
	}
	return 0;
}
//################################################################
//reads server public and private keys
int read_keys(const char *pub_key_path) {
	server_pub = read_key_file(pub_key_path);
	if (!server_pub) {
		return 0;
	}
	return 1;
}
//################################################################

//################################################################
//server info initialization
int init_server(int portnum) {
	memset(&server_addr, 0, sizeof(server_addr));
	server_addr.sin_family = AF_INET;
	server_addr.sin_addr.s_addr = INADDR_ANY;
	//server_addr.sin_port = htons(PWNC_SERVER_PORT);
	server_addr.sin_port = htons(portnum);

	sockfd = socket (PF_INET, SOCK_STREAM, 0);
	if (sockfd < 0) {
		fprintf(stderr, "init_server: socket creation failed\n");
		return 0;
	}

	//Bind a local address to the socket
	if (bind(sockfd, (struct sockaddr *)&server_addr, sizeof (server_addr)) < 0) {
		fprintf(stderr,"init_server: bind failed\n");
		return 0;
	}

	//Specify a size of request queue
	if (listen(sockfd, 20) < 0) {
		fprintf(stderr,"init_server: listen failed\n");
		return 0;
	}

	//init hash table
	clients_ht = lh_new(LHASH_HASH_FN(pwnc_client_hash), LHASH_COMP_FN(pwnc_client_cmp));
	if (!clients_ht) {
		return 0;
	}
	//init mutex
	pthread_mutex_init(&ht_mutex, NULL);

	return 1;
}
//################################################################

//################################################################
//Start firewall
//drops all incoming traffic coming from e.g. 192.168.1.0
//leaves open only PWNC_SERVER_PORT
void fw_init(const unsigned char *network, const unsigned char *mask) {
#ifndef FW_OFF
	unsigned char *command = NULL;
	ALLOC_AND_CLEAN(command, 256, unsigned char*);

	sprintf(command, "%s -N PWNC_INPUT", str_iptables_path);
	send_cmd(command);

	memset(command, 0, 256);
	sprintf(command, "%s -N PWNC_FORWARD", str_iptables_path);
	send_cmd(command);
//	system("/usr/sbin/iptables -N PWNC_INPUT");
//	system("iptables -N PWNC_FORWARD");

	sprintf(command, "%s -I INPUT 1 -s %s/%s -j PWNC_INPUT", str_iptables_path, network, mask);
	send_cmd(command);
	//system("iptables -I INPUT 1 -s 192.168.1.0/24 -j PWNC_INPUT");
	memset(command, 0, 256);
	sprintf(command, "%s -I FORWARD 1 -s %s/%s -j PWNC_FORWARD", str_iptables_path, network, mask);
	send_cmd(command);
	//system("iptables -I FORWARD 1 -s 192.168.1.0/24 -j PWNC_FORWARD");

	memset(command, 0, 256);
	sprintf(command, "%s -A PWNC_INPUT -s %s/%s --protocol tcp --destination-port 22 -j ACCEPT", str_iptables_path, network, mask);
	send_cmd(command);
	//permit messages on the PWNC_SERVER_PORT
	memset(command, 0, 256);
	sprintf(command, "%s -A PWNC_INPUT -s %s/%s --protocol tcp --destination-port %d -j ACCEPT", str_iptables_path, network, mask, PWNC_SERVER_PORT);
	send_cmd(command);

	//permit only ssh
//	system("iptables -A INPUT -s 192.168.1.0/24 --protocol tcp --destination-port 22 -j ACCEPT");
	memset(command, 0, 256);
	sprintf(command, "%s -A PWNC_INPUT -s %s/%s -j REJECT", str_iptables_path, network, mask);
	send_cmd(command);
	//system("iptables -A PWNC_INPUT -s 192.168.1.0/24 -j REJECT");
	memset(command, 0, 256);
	sprintf(command, "%s -A PWNC_FORWARD -s %s/%s -j REJECT", str_iptables_path, network, mask);
	send_cmd(command);
	//system("iptables -A PWNC_FORWARD -s 192.168.1.0/24 -j REJECT");

	//system("iptables -A PWNC_INPUT -s 192.168.1.0/24 --protocol tcp --destination-port 7777 -j ACCEPT");

	SAFE_FREE(command);
#endif
}
//################################################################

//################################################################
//cleanup iptables rules added by the program
//it is meant to be called when the program terminates
void fw_cleanup(const unsigned char *network, const unsigned char *mask) {
#ifndef FW_OFF
	unsigned char *command = NULL;
	ALLOC_AND_CLEAN(command, 256, unsigned char*);

	//first remove references to the PWNC_xxx chains
	sprintf(command, "%s -D INPUT -s %s/%s -j PWNC_INPUT", str_iptables_path, network, mask);
	send_cmd(command);

	memset(command, 0, 256);
	sprintf(command, "%s -D FORWARD -s %s/%s -j PWNC_FORWARD", str_iptables_path, network, mask);
	send_cmd(command);

	//then flush the chains
	memset(command, 0, 256);
	sprintf(command, "%s -F PWNC_INPUT", str_iptables_path, network, mask);
	send_cmd(command);

	memset(command, 0, 256);
	sprintf(command, "%s -F PWNC_FORWARD", str_iptables_path, network, mask);
	send_cmd(command);

	//then remove the chains themselves
	memset(command, 0, 256);
	sprintf(command, "%s -X PWNC_INPUT", str_iptables_path, network, mask);
	send_cmd(command);

	memset(command, 0, 256);
	sprintf(command, "%s -X PWNC_FORWARD", str_iptables_path, network, mask);
	send_cmd(command);

	SAFE_FREE(command);
#endif
}
//################################################################

//################################################################
//Opens ports for ipaddr
//(adds appropriate rules on top of the rule stack)
void fw_permit_access(const char* ipaddr) {
#ifndef FW_OFF
	char command[1024];
	fprintf(stderr, "Permitting access\n");
	bzero(&command, 1024);
	sprintf((char*)command, "%s -I PWNC_INPUT 1 -s %s -j ACCEPT", str_iptables_path, ipaddr);
	send_cmd(command);
	fprintf(stderr, "%s\n", command);

	bzero(&command, 1024);
	sprintf((char*)command, "%s -I PWNC_FORWARD 1 -s %s -j ACCEPT", str_iptables_path, ipaddr);
	send_cmd(command);
	fprintf(stderr, "%s\n", command);
#endif
}
//################################################################

//################################################################
//blocks access to ipaddr
void fw_block_access(const char* ipaddr) {
#ifndef FW_OFF
	fprintf(stderr, "blocking access\n");
	char command[1024];
	bzero(&command, 1024);
	sprintf((char*)command, "%s -D PWNC_INPUT -s %s -j ACCEPT", str_iptables_path, ipaddr);
	send_cmd(command);
	bzero(&command, 1024);
	sprintf((char*)command, "%s -D PWNC_FORWARD -s %s -j ACCEPT", str_iptables_path, ipaddr);
	send_cmd(command);
#endif
}
//################################################################

//################################################################
//Create /dev/tstats
int create_stats_dev() {
#ifndef _TLOG_OFF
	dev_t tstats_dev;
	//create a dev_t structure with major: 241, minor:0
	tstats_dev = makedev(241, 0);
	//create char devive with rw permissions
	mknod ("/dev/tstats", S_IFCHR | 0666, tstats_dev);
#endif
	return 0;
}
//################################################################

//################################################################
//Adds a client record to /dev/tstats
int add_tstats_rec(unsigned long int client_addr) {
	int retval = 0;
#ifndef _TLOG_OFF
	int f = open("/dev/tstats", O_RDWR);
	if (f == -1) {
		fprintf(stderr, "add_stats_rec: Error opening /dev/tstats\n");
		return -1;
	}
	retval = ioctl(f, ADD_REC,(void*)&(client_addr));
	close(f);
#endif
	return retval;
}
//################################################################

//################################################################
//Removes  a client record from /dev/tstats
int remove_tstats_rec(struct in_addr clientAddr) {
	int retval = 0;
#ifndef _TLOG_OFF
	int f = open("/dev/tstats", O_RDWR);

	retval = ioctl(f, REMOVE_REC,(void*)&(clientAddr.s_addr));
	close (f);
#endif
	return retval;
}
//################################################################

//################################################################
//Gets the traffic a client has generated (polls /dev/tstats)
unsigned int get_client_traffic(unsigned long int client_addr) {
	int retval = 0;
	unsigned int nBytes = 0;
#ifndef _TLOG_OFF
	int f = open("/dev/tstats", O_RDWR | O_EXCL);

//	struct stats* client_stats = (struct stats*)malloc(sizeof(struct stats));
//	bzero(client_stats, sizeof(struct stats));

	struct stats client_stats;
	memset(&client_stats, 0, sizeof(struct stats));

	client_stats.nClientIP = client_addr;
	retval = ioctl(f, GET_STATS,(void*)&client_stats);

//	if (client_stats) {
		nBytes += client_stats.nTotalBytesIn;
		nBytes += client_stats.nTotalBytesOut;
//	}
	close(f);

//	SAFE_FREE(client_stats);
	//return kilobytes
	//return (nBytes/1000);
#else
	nBytes = 100;
#endif
	return (nBytes);
}
//################################################################

//################################################################
int send_receipt_to_tca(PWNC_RCPT *rcpt) {
	//fprintf(stderr, "\nsend_receipt_to_tca\n");
	struct sockaddr_in tca_addr;
	int tca_sockfd;
	int alen = 0;
	unsigned char *text = NULL;
	int mtype = 0;
	int ecode = 0; //error code (or OK)
	int select_status = 0;
	fd_set fs;
	struct timeval delay;

	memset(&tca_addr, 0, sizeof(tca_addr));
	tca_addr.sin_family = AF_INET;
	inet_aton(str_tca_ipaddr, &(tca_addr.sin_addr) );
	tca_addr.sin_port = htons(PWNC_TCA_PORT);

	alen = sizeof(tca_addr);

	tca_sockfd = socket (PF_INET, SOCK_STREAM, 0);
	if (tca_sockfd < 0) {
		fprintf(stderr, "socket creation failed\n");
		return 0;
	}

	if ( connect (tca_sockfd, (const struct sockaddr*)&tca_addr, (socklen_t)alen) == -1) {
		fprintf(stderr, "Error connecting to TCA\n");
		close(tca_sockfd);
		return 0;
	}

	text = receipt_2_text(rcpt);
	if (!text) {
		fprintf(stderr, "Error constructing receipt message\n");
		close(tca_sockfd);
		return 0;
	}
	fprintf(stderr, "\nconstructed receipt\n");

	//send rcpt
	fprintf(stderr, "\nSending receipt to TCA\n");
	if (!write(tca_sockfd, text, strlen(text))) {
		fprintf(stderr, "Error sending receipt to TCA\n");
		close(tca_sockfd);
		SAFE_FREE(text)
		return 0;
	}

	SAFE_FREE(text);

	//wait at most 20 seconds till TCA answers
	//no answer ==> error
	FD_ZERO(&fs);
	FD_SET(tca_sockfd, &fs);

	memset(&delay, 0, sizeof(delay));
	delay.tv_sec = 20;//timeout: 20 seconds
	delay.tv_usec = 0;

	if ( (select_status = select(tca_sockfd + 1, &fs, NULL, NULL, &delay)) > 0) {
		text = read_message(tca_sockfd, &mtype);
		fprintf(stderr, "%s\n", text);
	}
	else {
		//timeout!!
		SAFE_FREE(text);
		close(tca_sockfd);
		return 0;
	}

	//now parse RRSP
	if (mtype != PWNC_MSG_RRSP) {
		//wrong message type
		SAFE_FREE(text);
		close(tca_sockfd);
		return 0;
	}

	ecode = parse_rrsp(text);
	SAFE_FREE(text);
	close(tca_sockfd);

	if (ecode != 1) {
		fprintf(stderr, "TCA receipt error: %d\n", ecode);
	}

	return ecode;
}

int query_tca(PWNC_PUBKEY *key) {
	struct sockaddr_in tca_addr;
	int tca_sockfd;
	int alen = 0;
	unsigned char *text = NULL;
	int mtype = 0;
	int select_status = 0;
	fd_set fs;
	struct timeval delay;

	memset(&tca_addr, 0, sizeof(tca_addr));
	tca_addr.sin_family = AF_INET;
	inet_aton(str_tca_ipaddr, &(tca_addr.sin_addr) );
	tca_addr.sin_port = htons(PWNC_TCA_PORT);

	alen = sizeof(tca_addr);

	tca_sockfd = socket (PF_INET, SOCK_STREAM, 0);
	if (tca_sockfd < 0) {
		fprintf(stderr, "socket creation failed\n");
		return 0;
	}

	if ( connect (tca_sockfd, (const struct sockaddr*)&tca_addr, (socklen_t)alen) == -1) {
		fprintf(stderr, "Error connecting to TCA\n");
		close(tca_sockfd);
		return 0;
	}

	//construct query to be sent to tca
	text = construct_query(server_pub, key);
	if (!text) {
		fprintf(stderr, "Error constructing TCA query\n");
		close(tca_sockfd);
		return 0;
		}

	//send query
	if (!write(tca_sockfd, text, strlen(text))) {
		fprintf(stderr, "Error sending query to TCA\n");
		close(tca_sockfd);
		SAFE_FREE(text)
		return 0;
	}

	SAFE_FREE(text);

	//now wait at most 20 seconds till TCA answers
	//no answer ==> some error, etc
	FD_ZERO(&fs);
	FD_SET(tca_sockfd, &fs);

	memset(&delay, 0, sizeof(delay));
	delay.tv_sec = 20;//timeout: 20 seconds
	delay.tv_usec = 0;

	if ( (select_status = select(tca_sockfd + 1, &fs, NULL, NULL, &delay)) > 0) {
		text = read_message(tca_sockfd, &mtype);
		fprintf(stderr, "%s\n", text);
	}
	else {
		//timeout!!
		SAFE_FREE(text);
		close(tca_sockfd);
		return 0;
	}

	//now parse QRSP
	if (mtype != PWNC_MSG_QRSP) {
		//wrong message type
		SAFE_FREE(text);
		close(tca_sockfd);
		return 0;
	}

	if (parse_qrsp(text) != 1) {
		fprintf(stderr, "User NOT granted\n");
		SAFE_FREE(text);
		close(tca_sockfd);
		return 0;
	}

	close(tca_sockfd);

	//user must be granted access
	return 1;
}
//################################################################

//################################################################
//Ends a client session
//Blocks access to client addess,
//removes from /dev/tsats
//removes from hash table
//TODO: sends receipt to TTP
//void end_client_session(int s) {
void end_client_session(int tindex, unsigned long client_addr) {
	struct sockaddr_in client;
	int addr_len = 0;
	unsigned char *straddr = NULL;
	PWNC_CLIENT *pclient = NULL;
	PWNC_CLIENT *to_delete = NULL; //temp client structure, used when deleting an entry from the hash table
	addr_len = sizeof(client);
	memset(&client, 0, addr_len);

	//fprintf(stderr, "END_CLIENT_SESSION\n");
	client.sin_addr.s_addr = client_addr;
	client.sin_family = AF_INET;
//	if (getpeername(s, (struct sockaddr*)&client, (socklen_t*)&addr_len) == -1) {
//		printf("%d - %s\n", errno, strerror(errno));
//		printf("end_client_session: Error extracting addr info\n");
//		pthread_exit(NULL);
//	}

	ALLOC_AND_CLEAN(straddr, INET_ADDRSTRLEN, unsigned char*);
	if (!inet_ntop(AF_INET, (const void*)&(client.sin_addr), straddr, INET_ADDRSTRLEN)) {
		fprintf(stderr, "Error converting address to dotted decimal form\n");
		pthread_exit(NULL);
	}

	fw_block_access(straddr);
	//remove record from /dev/tsats
	remove_tstats_rec(client.sin_addr);

	//first retrieve client from hash table
	//TODO: then, free allocated memory ??? (check it)
	//finally, remove hash table entry
	pthread_mutex_lock(&ht_mutex);
	to_delete = PWNC_CLIENT_new(client.sin_addr.s_addr);
	pclient = (PWNC_CLIENT*)lh_delete(clients_ht, (void*)to_delete); //deleted from hash table
	SAFE_FREE_CLIENT(to_delete);
	if (pclient) {
		//send rcpt to tca
#ifdef TCA_SUPPORT
		if (strlen(str_tca_ipaddr)) {
			if (pclient->rcpt) {
				if (send_receipt_to_tca(pclient->rcpt) != 1) {
					//TODO: do sth with the receipt, make checks, etc
				}
			}
		}
#endif
		//finally, free mem allocated for the deleted client
		SAFE_FREE_CLIENT(pclient);
	}
	pthread_mutex_unlock(&ht_mutex);

//	close(s);
	exit_client_thread(tindex);
	//pthread_exit(NULL);
}
//################################################################

//################################################################
//void client_handler (int sockfd) {
void client_handler(THREAD_INFO *ti) {
	unsigned char *text = NULL; //text message
	int mtype = 0; //message type
	PWNC_CERT *client_cert; //client certificate
//	PWNC_RCPT *rcpt; //the most recent valid receipt the server has received from the client
	PWNC_RCPT *rcpt_temp; //the last receipt received (not yet checked for validity, etc)
	PWNC_RCPT *rcpt_temp_2; //used only when swapping the above receipts;
	unsigned long weight = 0; //receipt weight (it will be placed on an rreq msg)
	unsigned char *cert_info; //serialized certificate info
	unsigned char *rinfo; //serialized receipt info
	int info_len = 0; //cert info (serialized) length
	int verified = 0;
	PWNC_CLIENT *pclient = NULL;
	PWNC_CLIENT *to_retrieve = NULL; //temp client struct, used to retrieve records from the hash table
	struct timeval delay; //delay
	int select_status = 0; //used in "select" calls
	fd_set fs;

	//client threads ignore SIGINT and SIGKILL
	//these signals are handled by the main thread, which
	//first cancels client threads and then cleans up and exits
	if (block_signals()) {
		fprintf(stderr, "Warning: Could not set signal mask\n");
	}

	//get thread info
	int sockfd = ti->client_socket;
	int tindex = ti->thread_index;
	SAFE_FREE(ti);

	struct sockaddr_in client_addr;
	unsigned char *straddr = NULL; //string representation of ip address of client
	int addr_len = sizeof(client_addr);
	memset(&client_addr, 0, addr_len);

	//get client address from socket
	if (getpeername(sockfd, (struct sockaddr*)&client_addr, (socklen_t*)&addr_len) == -1) {
		fprintf(stderr, "client_handler: Error extracting addr info\n");
		exit_client_thread(tindex); //exit thread
		return;
	}
	ALLOC_AND_CLEAN(straddr, INET_ADDRSTRLEN, unsigned char*);
	if (!inet_ntop(AF_INET, (const void*)&(client_addr.sin_addr), straddr, INET_ADDRSTRLEN)) {
		fprintf(stderr, "Error converting address to dotted decimal form\n");
		exit_client_thread(tindex);
		return;
	}

	//read message from socket
	FD_ZERO(&fs);
	FD_SET(sockfd, &fs);
	delay.tv_sec = 10L;//timeout: 10 sec
	delay.tv_usec = 0L;
	if ( (select_status = select(sockfd + 1, &fs, NULL, NULL, &delay)) > 0) {
		text = read_message(sockfd, &mtype);
		if (!text) {
			close(sockfd);
			exit_client_thread(tindex);
			return;
		}
		fprintf(stderr, "\n%s\n\n", text);
	}
	else {
		//select timeout - close connection -- exit
		close(sockfd);
		exit_client_thread(tindex);
		return;
	}

	PWNC_RCPT *rcpt = NULL; //used when client starts session with RCPT
	if (mtype == PWNC_MSG_RCPT) {
		//if the client starts the session with a receipt
		//we insert the receipts and wait for a conn msg
		//if (timeout) --> end client session
		do {
			rcpt = text_2_rcpt(text);

//TODO: TEAM_SERVER, TCA_SUPPORT mutually exclusive????
#ifdef TEAM_SERVER
//TODO: do stuff with receipt (send to team server, etc)
#endif

#ifdef TCA_SUPPORT
			//send receipt to TCA
			//do not perform verifications, rdb_engine will do so, just send it...
			if (strlen(str_tca_ipaddr)) {
				if (rcpt) {
					if (send_receipt_to_tca(rcpt) != 1) {
						//TODO: do sth with the receipt, make checks, etc
					}
				}
			}
#endif
			SAFE_FREE_RCPT(rcpt); //free receipt
			SAFE_FREE(text);//free text
//TODO++ select on socket ++TODO
			text = read_message(sockfd, &mtype); //read next message from client
			//...and loop
		} while (mtype == PWNC_MSG_RCPT); //if !true, exits loop and checks if message is CONN (now the real session starts...), else loop...
	}

	if (mtype == PWNC_MSG_CONN) {
		//extract client certificate
		client_cert = parse_conn_msg(text);
		SAFE_FREE(text);
		if (!client_cert) {
			fprintf(stderr, "Error parsing CONN message\n");
			exit_client_thread(tindex);
		}
		//verify certificate
		//first serialize cert info and then verify signature
		cert_info = serialize_cert_info(client_cert->TeamPubKey, client_cert->UserPubKey, &info_len);
		if (!cert_info) {
			//err
			fprintf(stderr, "Error serializing certificate info\n");
			exit_client_thread(tindex);
		}
		verified = verify_pwnc_signature(client_cert->TeamSig, client_cert->TeamPubKey, cert_info, info_len, 1);
		if (verified != 1) {
			//remove user
			fprintf(stderr, "Error verifying client cert\n");
			exit_client_thread(tindex);
		}

		//now server has to ask the TCA if the client should be admitted
		//if not, send err message
		//if its OK goto next step (create state, etc....) --->
#ifdef TCA_SUPPORT
		if (strlen(str_tca_ipaddr) && !patient) {
			//if the ap is in patience mode
			//it allows free access, so it doesn't query the
			//repository server
			if (query_tca(client_cert->TeamPubKey)) {
				fprintf(stderr, "Client admitted\n");
			}
			else {
				//should not be admitted, kill thread, close socket, etc
				//close(sockfd);
				exit_client_thread(tindex);
				return;
			}
		}
#endif

		//so far so good...
		//create new client entry (extract client pubkey, insert client into hash table, etc)
		//open ports, etc

		//TODO: check if client exists and ignore message(?)
		pthread_mutex_lock(&ht_mutex);
		to_retrieve = PWNC_CLIENT_new(client_addr.sin_addr.s_addr);
		pclient = (PWNC_CLIENT*)lh_retrieve(clients_ht, (void*)to_retrieve);
		SAFE_FREE_CLIENT(to_retrieve);
		if (!pclient) {
			//fprintf(stderr, "constructing pclient\n");
			//new client
			pclient = PWNC_CLIENT_new(client_addr.sin_addr.s_addr);
			//copy client_cert to pclient
			pclient->client_cert->TeamPubKey->algorithm = client_cert->TeamPubKey->algorithm;
			pclient->client_cert->TeamPubKey->bits = client_cert->TeamPubKey->bits;
			pclient->client_cert->TeamPubKey->datalen = client_cert->TeamPubKey->datalen;
			ALLOC_AND_CLEAN(pclient->client_cert->TeamPubKey->data, pclient->client_cert->TeamPubKey->datalen, unsigned char*);
			memcpy(pclient->client_cert->TeamPubKey->data, client_cert->TeamPubKey->data, client_cert->TeamPubKey->datalen);

			pclient->client_cert->UserPubKey->algorithm = client_cert->UserPubKey->algorithm;
			pclient->client_cert->UserPubKey->bits = client_cert->UserPubKey->bits;
			pclient->client_cert->UserPubKey->datalen = client_cert->UserPubKey->datalen;
			ALLOC_AND_CLEAN(pclient->client_cert->UserPubKey->data, pclient->client_cert->UserPubKey->datalen, unsigned char*);
			memcpy(pclient->client_cert->UserPubKey->data, client_cert->UserPubKey->data, client_cert->UserPubKey->datalen);

			pclient->client_cert->TeamSig->algorithm = client_cert->TeamSig->algorithm;
			pclient->client_cert->TeamSig->bits = client_cert->TeamSig->bits;
			pclient->client_cert->TeamSig->datalen = client_cert->TeamSig->datalen;
			ALLOC_AND_CLEAN(pclient->client_cert->TeamSig->data, pclient->client_cert->TeamSig->datalen, unsigned char*);
			memcpy(pclient->client_cert->TeamSig->data, client_cert->TeamSig->data, client_cert->TeamSig->datalen);
			pclient->sockfd = sockfd;
			pclient->timestamp = (unsigned long)time(NULL);
			//lh_insert client
			lh_insert(clients_ht, (void*)pclient);
			//unlock hash table
			//fprintf(stderr, "Client inserted into hash table\n");
		}
		else {
			//else ignore message - some error occured (client already exists)
			fprintf(stderr, "Client already in the hash table\n");
			//close (sockfd);
			pthread_mutex_unlock(&ht_mutex);
			exit_client_thread(tindex);
			return;
		}

		pthread_mutex_unlock(&ht_mutex);

		//start traffic logging
		if (add_tstats_rec(client_addr.sin_addr.s_addr) == -1) {
			fprintf(stderr, "Failed to create char dev entry\n");
			//remove user from hash table
			pthread_mutex_lock(&ht_mutex);
			to_retrieve = PWNC_CLIENT_new(client_addr.sin_addr.s_addr);
			pclient = (PWNC_CLIENT*)lh_delete(clients_ht, (void*)to_retrieve);
			SAFE_FREE_CLIENT(to_retrieve);
			if (pclient) {
				SAFE_FREE_CLIENT(pclient);
			}
			pthread_mutex_unlock(&ht_mutex);
			exit_client_thread(tindex);
			return;
		}

		//open ports (which were blocked)
		fw_permit_access(straddr);
		SAFE_FREE(straddr);
		
		//reply to client(cack)
		fprintf(stderr, "Constructing CACK message\n");
		text = construct_conn_ack((unsigned long)time(NULL), MODE_RCPT_REQ_OFF);
		if (!text) {
			//error
			fprintf(stderr, "Error constructing CACK message\n");
			//end_client_session(sockfd);
			end_client_session(tindex, client_addr.sin_addr.s_addr);
			return;
		}

		//send CACK to client
		fprintf(stderr, "About to write CACK message to socket\n");
		if (!write(sockfd, text, strlen(text))) {
			//error writing to client socket
			//remove state, close ports, etc...
			//end_client_session(sockfd);
			SAFE_FREE(text);
			end_client_session(tindex, client_addr.sin_addr.s_addr);
			return;
		}
		SAFE_FREE(text);

/*
		//TODO: ayto to eixa kanonika alla to evgala twra  -allages pou kanw gia na metrhsw rreq/sec, etc
		//now sleep for a few secs and then sent RREQ
		delay.tv_sec = 3;
		delay.tv_usec = 0;
		select(0, NULL, NULL, NULL, &delay);
*/

		//calculate the amount of traffic the client has initiated
		weight = (unsigned long)get_client_traffic(client_addr.sin_addr.s_addr);
		fprintf(stderr, "Client traffic: %ld\n", weight);

		text = construct_rreq(server_pub, weight);
		fprintf(stderr, "Sending RREQ msg to client\n");
		if (!write(sockfd, text, strlen(text))) {
			fprintf(stderr, "Error writing to socket\n");
			//end_client_session(sockfd);
			SAFE_FREE(text);
			end_client_session(tindex, client_addr.sin_addr.s_addr);
			return;
		}
		//fprintf(stderr, "written to socket\n");
		SAFE_FREE(text);
	}
	else {
		//we expect that the first client message is CONN
		//if not, reject client
		fprintf(stderr, "Error - invalid message type\n");
		close(sockfd);
		exit_client_thread(tindex);
	}

	while (1) {
		FD_ZERO(&fs);
		FD_SET(sockfd, &fs);
		delay.tv_sec = (long)RREQ_DELAY_SEC;//timeout: rreq_interval seconds
		delay.tv_usec = (long)RREQ_DELAY_USEC;
		if ( (select_status = select(sockfd + 1, &fs, NULL, NULL, &delay)) > 0) {
			text = read_message(sockfd, &mtype);
			if (!text) {
				//end_client_session(sockfd);
				end_client_session(tindex, client_addr.sin_addr.s_addr);
				pthread_exit(NULL);
			}
			fprintf(stderr, "%s\n", text);
		}
		else {
			//end_client_session(sockfd);
			end_client_session(tindex, client_addr.sin_addr.s_addr);
			pthread_exit(NULL);
		}

		switch(mtype) {
			case PWNC_MSG_RCPT: {
				if (!text) {
					//end_client_session(sockfd);
					end_client_session(tindex, client_addr.sin_addr.s_addr);
				}
				rcpt_temp = text_2_rcpt(text);
				SAFE_FREE(text);

				if (!rcpt_temp) {
					//error
					fprintf(stderr, "Error receiving receipt\n");
					//end_client_session(sockfd);
					end_client_session(tindex, client_addr.sin_addr.s_addr);
					return;
				}
				//First, retrieve client struct from hash table
				pthread_mutex_lock(&ht_mutex);
				to_retrieve = PWNC_CLIENT_new(client_addr.sin_addr.s_addr);
				pclient = (PWNC_CLIENT*)lh_retrieve(clients_ht, (void*)to_retrieve);
				SAFE_FREE_CLIENT(to_retrieve);
				pthread_mutex_unlock(&ht_mutex);

				if (!pclient) {
					//the client does not exist in the hash table
					//TODO: ti kanoume s'ayth thn periptwsh???kratame to receipt??ti exoume na xasoume....
					//end_client_session(sockfd);
					end_client_session(tindex, client_addr.sin_addr.s_addr);
					break;
				}

				//check receipt validity
				//certificate check
				//we check if the cert is the same as the one we already have
				//this way we don't have to verify it again (which would be time consuming)
				if (
					(pclient->client_cert->UserPubKey->datalen != rcpt_temp->ConsumerCert->UserPubKey->datalen) ||
					(pclient->client_cert->TeamPubKey->datalen != rcpt_temp->ConsumerCert->TeamPubKey->datalen) ||
					(pclient->client_cert->TeamSig->datalen != rcpt_temp->ConsumerCert->TeamSig->datalen) ||
					memcmp(pclient->client_cert->UserPubKey->data, rcpt_temp->ConsumerCert->UserPubKey->data, pclient->client_cert->UserPubKey->datalen) ||
					memcmp(pclient->client_cert->TeamPubKey->data, rcpt_temp->ConsumerCert->TeamPubKey->data, pclient->client_cert->UserPubKey->datalen) ||
					memcmp(pclient->client_cert->TeamSig->data, rcpt_temp->ConsumerCert->TeamSig->data, pclient->client_cert->TeamSig->datalen)
				) {
					//client certificate error...
					//end_client_session(sockfd);
					fprintf(stderr, "Client certificate error\n");
					end_client_session(tindex, client_addr.sin_addr.s_addr);
					return;
					//remove state, etc
				}

				//1st check: the receipt ProviderPubKey must be the same as ours
				if (
					(server_pub->datalen != rcpt_temp->ProviderPubKey->datalen) ||
					memcmp(server_pub->data, rcpt_temp->ProviderPubKey->data, rcpt_temp->ProviderPubKey->datalen)
				) {
					//error comparing server(provider) public key
					//remove state, close fw,etc
					//end_client_session(sockfd);
					fprintf(stderr, "Provider public key error\n");
					end_client_session(tindex, client_addr.sin_addr.s_addr);
					return;
				}

				//2nd check: receipt timestamp
				if (rcpt_temp->timestamp != pclient->timestamp) {
					//fprintf(stderr, "%ld - %ld \n", rcpt_temp->timestamp, pclient->timestamp);
					//end_client_session(sockfd);
					end_client_session(tindex, client_addr.sin_addr.s_addr);
					return;
				}

				//3nd check: verification of receipt signature
				rinfo = serialize_receipt_info(rcpt_temp->ConsumerCert, rcpt_temp->ProviderPubKey, rcpt_temp->timestamp, rcpt_temp->weight, &info_len);
				if (!rinfo) {
					//err
					fprintf(stderr, "Error serializing receipt info\n");
					//end_client_session(sockfd);
					end_client_session(tindex, client_addr.sin_addr.s_addr);
					return;
				}
				//TODO: edw to rcpt->ConsumerCert->UserPubKey prepei na antikatasta8ei me to pwnc_client->pubkey???

				verified = verify_pwnc_signature(rcpt_temp->RcptSig, client_cert->UserPubKey, rinfo, info_len, 1);

				SAFE_FREE(rinfo);

				if (verified != 1) {
					fprintf(stderr, "Receipt verification failed\n");
					//end_client_session(sockfd);
					end_client_session(tindex, client_addr.sin_addr.s_addr);
					return;
				}

				//if all checks are passed successfully, swap(rcpt, rcpt_temp)
				rcpt_temp_2 = pclient->rcpt;
				if (rcpt_temp_2) {
					pclient->rcpt = rcpt_temp;
					SAFE_FREE(rcpt_temp_2);
				}
				else {
					pclient->rcpt = rcpt_temp;
				}
				fprintf(stderr, "RECEIPT OK\n");

				//unlock mutex
				//pthread_mutex_unlock(&ht_mutex);

				//TODO: rate sensitive delay

				delay.tv_sec = (long)RREQ_DELAY_SEC;
				delay.tv_usec = (long)RREQ_DELAY_USEC;

				select(0, NULL, NULL, NULL, &delay);

				weight = (unsigned long)get_client_traffic(client_addr.sin_addr.s_addr);
				fprintf(stderr, "Client traffic: %ld\n", weight);
				text = construct_rreq(server_pub, weight);

				fprintf(stderr, "Sending RREQ msg to client\n");
				if (!write(sockfd, text, strlen(text))) {
					fprintf(stderr, "Error writing to socket\n");
					//end_client_session(sockfd);
					SAFE_FREE(text);
					end_client_session(tindex, client_addr.sin_addr.s_addr);
				}
				SAFE_FREE(text);
				break;
			}//case
			case PWNC_MSG_TEAR: {
				//this has to go...
				//close socket, send rcpt to ttp, remove client state
				break;
			}
			default: {
				//???
				break;
			}
		}
	}
}
//################################################################


//################################################################
void term_handler(int signum) {
	int i = 0;
	switch(signum) {
		case SIGINT: case SIGTERM: case SIGKILL:
			for (i=0; i<max_threads; i++) {
				cancel_thread(i);
			}
			fw_cleanup(str_subnet, str_mask);

			cleanup_ossl_threads();
			thread_pool_cleanup();
			exit(0);
	}
}
//################################################################


//################################################################
void print_usage() {
	fprintf(stderr, "\nUsage: ap ");
	fprintf(stderr, "[-h/--help] ");
	fprintf(stderr, "[-c/--conf file] ");
	fprintf(stderr, "[-pk/--public-key-path file] ");
	fprintf(stderr, "[-sp/--server-port port] ");
	fprintf(stderr, "[-tca/--tca-addr ip_address] ");
	fprintf(stderr, "[-mq/--mq-key message_queue_identifier] ");
	fprintf(stderr, "[-ipt/--iptables-path path] ");
	fprintf(stderr, "[-n/--subnet subnet_address] ");
	fprintf(stderr, "[-m/--mask subnet_mask] ");
	fprintf(stderr, "[-mt/--max-threads number_of_threads] ");
	fprintf(stderr, "[-rreq/--rreq-interval seconds] ");
	fprintf(stderr, "[-pt/--patient]\n\n");

	fprintf(stderr, "\t-h --help\t\tDisplay this message\n");
	fprintf(stderr, "\t-c --conf\t\tap program configuration file\n");
	fprintf(stderr, "\t-pk --public-key-path\tTeam public key file\n");
	fprintf(stderr, "\t-sp --server-port\tServer port\n");
	fprintf(stderr, "\t-tca --tca-addr\t\tTCA/Team server IP address\n");
	fprintf(stderr, "\t-mq --mq-key\t\tMessage queue identifier (for communicating with the cmd_handler process)\n");
	fprintf(stderr, "\t-ipt --ipt-path\t\tPath to the iptables program (usually /sbin/iptables)\n");
	fprintf(stderr, "\t-n --subnet\t\tLocal subnet (in dotted decimal form)\n");
	fprintf(stderr, "\t-m --mask\t\tSubnet mask (as an integer)\n");
	fprintf(stderr, "\t-mt --max-threads\tMaximum number of threads\n");
	fprintf(stderr, "\t-rreq --rreq-interval\tTime interval between successive RREQ (in seconds)\n");
	fprintf(stderr, "\t-pt --patient\t\tThe access point operates in \"patience\" mode (gives free access)\n\n");
}
//################################################################


//################################################################
int main(int argc, char** argv) {
	struct sockaddr_in client_addr;
	int sockfd2;
	int alen = 0;
	FILE *conf_file_fp = NULL;

	/*
	configuration
	*/
	memset(&str_tca_ipaddr, 0, 20);
	memset(&str_pub_key_path, 0, 80);
//	memset(&str_priv_key_path, 0, 80);
	memset(&str_conf_file_path, 0, 80);
	memset(&str_iptables_path, 0, 80);
	memset(&str_subnet, 0, 80);
	memset(&str_mask, 0, 4);

	int help = 0;

	CMD_ARG *cmd_line_args = NULL;
	CENTRY *conf_entries = NULL;
	unsigned char *valid_conf_names[] = {
		"public-key-path",
		"server-port",
		"tca-addr",
		"mq-key",
		"iptables-path",
		"subnet",
		"mask",
		"max-threads",
		"rreq-interval",
		"patient"
		};

	unsigned char *valid_cmd_line_names[] ={
		"conf",
		"public-key-path",
		"server-port",
		"tca-addr",
		"mq-key",
		"iptables-path",
		"subnet",
		"mask",
		"max-threads",
		"rreq-interval",
		"patient",
		"help"
		};

	unsigned char *cmd_line_switches[] = {"c", "pk", "sp", "tca", "mq", "ipt", "n", "m", "mt", "rreq", "pt", "h"};

	unsigned short conf_var_types[] = {
		VTYPE_STRING, //public key
		VTYPE_INT, //server port
		VTYPE_STRING, //tca addr
		VTYPE_INT, //mq key
		VTYPE_STRING, //iptables path
		VTYPE_STRING, //subnet
		VTYPE_STRING, //subnet mask
		VTYPE_INT, //max threads
		VTYPE_INT, //rreq interval
		VTYPE_NOVAL //patient
		};

	unsigned short cmd_line_var_types[] = {
		VTYPE_STRING, //conffile
		VTYPE_STRING, //pubkey
		VTYPE_INT, //server port
		VTYPE_STRING, //tca addr
		VTYPE_INT, //mq key
		VTYPE_STRING, //iptables path
		VTYPE_STRING, //subnet
		VTYPE_STRING, //mask
		VTYPE_INT, //max threads
		VTYPE_INT, //rreq
		VTYPE_NOVAL, //patient
		VTYPE_NOVAL //help
		};

	void *conf_variables[] = {
		&str_pub_key_path,
		&server_port,
		&str_tca_ipaddr,
		&mq_key,
		&str_iptables_path,
		&str_subnet,
		&str_mask,
		&max_threads,
		&RREQ_DELAY_SEC,
		&patient
		};

	void *cmd_line_variables[] = {
		&str_conf_file_path,
		&str_pub_key_path,
		&server_port,
		&str_tca_ipaddr,
		&mq_key,
		&str_iptables_path,
		&str_subnet,
		&str_mask,
		&max_threads,
		&RREQ_DELAY_SEC,
		&patient,
		&help
		};

	conf_entries = make_valid_entries(valid_conf_names, conf_variables, conf_var_types, 10);
	cmd_line_args = make_valid_args(valid_cmd_line_names, cmd_line_switches, cmd_line_variables, cmd_line_var_types, 12);

	if (argc < 2) {
		//no command line flags specified, so parse default conf file
		conf_file_fp = fopen(DEF_CONF_PATH, "r");
		if (!conf_file_fp) {
			//first check in /etc/pwnc.conf
			conf_file_fp = fopen(DEF_CONF_FNAME, "r");
			if (!conf_file_fp) {
				//if fopen failed, try to open pwnc.conf in the current path
				//if all else fails ---> exit
				fprintf(stderr, "ap: Error opening configuration file\n\n");
				print_usage();
				exit(1);
			}
		}
		//default conffile opened ok, so parse it
		parse_conf_file(conf_entries, 10, conf_file_fp);
	}
	else {
		//parse command line
		//if there is a -c flag, ignore all other flags and check the specified conf file
		parse_cmd_line_args(argc, argv, cmd_line_args, 12);
		if (help) {
			//user gave the -h flag (probably among others...)
			print_usage();
			exit(0);
		}
		if (strlen(str_conf_file_path)) {
			//there was a -c flag, so ignore all other flags and check this conf file
			conf_file_fp = fopen(str_conf_file_path, "r");
			if (!conf_file_fp) {
				fprintf(stderr, "ap: Warning: Failed to open %s, trying defaults\n", str_conf_file_path);
				//try to open specified conffile - if it fails, try defaults again
				conf_file_fp = fopen(DEF_CONF_PATH, "r");
				if (!conf_file_fp) {
					//trying default conffile path
					conf_file_fp = fopen(DEF_CONF_FNAME, "r");
					if (!conf_file_fp) {
						//trying to open default conf file in current dir
						fprintf(stderr, "ap: Error opening configuration file\n\n");
						print_usage();
						exit(1);
					}
				}
			}
			parse_conf_file(conf_entries, 10, conf_file_fp);
		}
		//otherwise: we are using cmd line arguments
	}

	//check mandatory fields (keys, mq key, ipt path)
	if (!strlen(str_pub_key_path) || !mq_key || !strlen(str_iptables_path) || !server_port || !strlen(str_subnet) || !strlen(str_mask)) {
		fprintf(stderr, "ap: Some configuration info is missing\n\n");
		print_usage();
		exit(1);
	}

	fprintf(stderr, "Access point configuration:\n");
	fprintf(stderr, "---------------------------\n");
	fprintf(stderr, "Public key path: %s\n", str_pub_key_path);
	fprintf(stderr, "Repository (TCA) IP address: %s\n", str_tca_ipaddr);
	fprintf(stderr, "Listening to port: %d\n", server_port);
	fprintf(stderr, "Message queue key: %d\n", mq_key);
	fprintf(stderr, "Path to the iptables executable: %s\n", str_iptables_path);
	fprintf(stderr, "Subnet: %s\n", str_subnet);
	fprintf(stderr, "Subnet mask: %s\n", str_mask);
	fprintf(stderr, "Maximum number of threads: %d\n", max_threads);
	fprintf(stderr, "RREQ interval: %d\n", RREQ_DELAY_SEC);
	fprintf(stderr, "Patience mode: %s\n\n", (patient?"ON":"OFF") );

	signal(SIGINT, term_handler);
	signal(SIGTERM, term_handler);
	signal(SIGKILL, term_handler);

	//initialize message queue
	if (init_mq() == -1) {
		fprintf(stderr, "ap: Could not initialize message queue\n");
		exit(1);
	}

	if (!read_keys(str_pub_key_path)) {
		fprintf(stderr, "ap: Error reading keys\n");
		exit(1);
	}

	//init firewall
	fw_init(str_subnet, str_mask);

	create_stats_dev();

	if (!init_server(server_port)) {
		SAFE_FREE_KEY(server_pub);
		fw_cleanup(str_subnet, str_mask);
		exit(1);
	}

	alen = sizeof(client_addr);

	//seed prng
	init_prng();

	//init openssl mt support
	init_ossl_threads();

	//initialize thread pool and thread pool mutex
	init_thread_pool(max_threads);
	pthread_mutex_init(&thread_pool_mutex, NULL);

	//daemonize process...
	daemonize();

	while (1) {
		fprintf(stderr, "ap: Waiting for client connections...\n");

		if (  (sockfd2=accept(sockfd, (struct sockaddr *)&client_addr, &alen)) < 0) {
			fprintf(stderr, "ap: accept failed\n");
			continue;
//			exit (1);
		}
		start_new_thread((void*)&client_handler, sockfd2);
	}
	close(sockfd);

	return 0;
}
//################################################################

void daemonize() {
	int fd = 0;

	if((fd = open("/dev/null", O_WRONLY)) == -1) {
		fprintf(stderr, "ap: Cannot open /dev/null\n");
	}
	else {
		close(STDIN_FILENO);
		dup2(fd, STDOUT_FILENO);
		dup2(fd, STDERR_FILENO);
	}

	if(fork()) {
		exit(0);
	}
	setsid();
}
